import java.awt.*;
import java.io.*;
import java.awt.event.*;
import java.util.regex.*;
import ij.*;
import ij.io.*;
import ij.gui.*;
import ij.text.*;
import ij.plugin.filter.Analyzer;

/** Author: Jon Jackson j.jackson@ucl.ac.uk
*
*   Last Modified 01/09/05
*
*   Uses a TextPanel to displays text in a window.
*   Adds Filter Panel to filter displayed fields
*   
*   Code is an extension of ImageJ's TextWindow
*
*   Filter assumes that the text panel contains two columns (Field and Value)
*   The columns are assumed to be tab separated, and entries new line new line separated
*   Some aspects of the filter may not work if text is not in this format
*/

public class SearchWindow implements ActionListener, ItemListener, FocusListener, WindowListener {
    Frame frame;
    TextPanel textPanel;
    Panel utilityPanel;
    Panel mainPanel;
    TextField filterText;
    Choice filterChoice;
	Choice outputChoice;
	Choice separatorChoice;
    String allData;
    MenuBar mb;
    String aboutText="Output button sends all (or selected) lines to ImageJ log window\n" + 
				"or UNIX Terminal with the given separator.\n \n" +
				"Filter button displays entries matching a given pattern\n \n" +
				"'Starts With' and 'Ends With' options search only 'Field' entries\n" +
				"'Regular Expression' searches entire line (column separator is \\t)\n \n" +
				"Regular Expression syntax can be found at \n" +
				"http://java.sun.com/j2se/1.5.0/docs/api/java/util/regex/Pattern.html\n \n" +
				"Feedback to j.jackson@ucl.ac.uk";
    /**
    Opens a new single-column text search window.
    @param title    the title of the window
    @param str        the text initially displayed in the window
    @param width    the width of the window in pixels
    @param height    the height of the window in pixels
    */
    public SearchWindow(String title, String data, int width, int height) {
        this(title, "", data, width, height);
    }
    
    /**
    Opens a new multi-column text search window.
    @param title    the title of the window
    @param headings    the tab-delimited column headings
    @param data        the text initially displayed in the window
    @param width    the width of the window in pixels
    @param height    the height of the window in pixels
    */
    public SearchWindow(String title, String headings, String data, int width, int height) {
        frame = new Frame(title);
        allData = data;
        addMainPanel(title, headings);
        if (data!=null && !data.equals(""))
            textPanel.append(data);
        frame.addKeyListener(textPanel);
        ImageJ ij = IJ.getInstance();
        if (ij!=null) {
            Image img = ij.getIconImage();
            if (img!=null)
                try {frame.setIconImage(img);} catch (Exception e) {}
        }
        frame.addFocusListener(this);
        frame.addWindowListener(this);
        addMenuBar();
        WindowManager.addWindow(frame);
        frame.setSize(width, height);
        GUI.center(frame);
        frame.setVisible(true);
    }

    /**
    Opens a new text search window containing the contents
    of a text file.
    @param path        the path to the text file
    @param width    the width of the window in pixels
    @param height    the height of the window in pixels
    */
    public SearchWindow(String path, int width, int height) {
        frame = new Frame();
        addMainPanel("", "");
        if (openFile(path)) {
            frame.addFocusListener(this);
            WindowManager.addWindow(frame);
            frame.setSize(width, height);
            frame.setVisible(true);
            allData = textPanel.getText();
        } else
            frame.dispose();
    }
    
    /** Clear contents of SearchWindow and insert new text
    */
    public void setText(String text) {
        if (text == null)
            return;
        allData = text;
        textPanel.selectAll();
        textPanel.clearSelection();
        textPanel.append(allData);
    }
    
    void addMainPanel(String title, String headings) {
        textPanel = new TextPanel(title);
        textPanel.setTitle(title);
        textPanel.setColumnHeadings(headings);
        utilityPanel = new Panel();
        utilityPanel.setLayout(new BorderLayout());
        Panel buttonPanel = new Panel();
		Button outputButton = new Button("Output");
        outputButton.addActionListener(this);
        buttonPanel.add(outputButton);
		outputChoice = new Choice();
		outputChoice.add("Log");
		outputChoice.add("Terminal");
		buttonPanel.add(outputChoice);
		separatorChoice = new Choice();
		separatorChoice.add("Tab");
		separatorChoice.add("Space");
		separatorChoice.add("Equals");
		buttonPanel.add(separatorChoice);
        utilityPanel.add(buttonPanel, BorderLayout.NORTH);
        if (IJ.isJava14()) {
            Panel filterPanel = new Panel();
            Button filterButton = new Button("Filter");
            filterButton.addActionListener(this);
            filterPanel.add(filterButton);
            filterText = new TextField("", 15);
		    filterText.addActionListener(this);
            filterPanel.add(filterText);
            filterChoice = new Choice();
            filterChoice.add("Starts With");
            filterChoice.add("Field Contains");
            filterChoice.add("Value Contains");
            filterChoice.add("Ends With");
            filterChoice.add("Regular Expression");
            filterChoice.select("Contains");
		    filterChoice.addItemListener(this);
            filterPanel.add(filterChoice);
            Button allButton = new Button("Reset");
            allButton.addActionListener(this);
            filterPanel.add(allButton);
            utilityPanel.add(filterPanel, BorderLayout.CENTER);
        } else {
            IJ.showStatus("Search facility requires Java 1.4");
        }
        mainPanel = new Panel();
        mainPanel.setLayout(new BorderLayout());
        mainPanel.add(textPanel, BorderLayout.CENTER);
        mainPanel.add(utilityPanel, BorderLayout.SOUTH);
        frame.add(mainPanel);
    }
    
    void addMenuBar() {
        mb = new MenuBar();
        Menu m = new Menu("File");
        m.add(new MenuItem("Save As..."/*, new MenuShortcut(KeyEvent.VK_S)*/));
        m.addActionListener(this);
        mb.add(m);
        m = new Menu("Edit");
        m.add(new MenuItem("Cut", new MenuShortcut(KeyEvent.VK_X)));
        m.add(new MenuItem("Copy", new MenuShortcut(KeyEvent.VK_C)));
        m.add(new MenuItem("Copy All"));
        m.add(new MenuItem("Clear"));
        m.add(new MenuItem("Select All", new MenuShortcut(KeyEvent.VK_A)));
		m.add(new MenuItem("Clear Filter", new MenuShortcut(KeyEvent.VK_D)));
        if (frame.getTitle().equals("Results")) {
            m.addSeparator();
            m.add(new MenuItem("Clear Results"));
            m.add(new MenuItem("Summarize"));
            m.add(new MenuItem("Set Measurements..."));
        }
        m.addActionListener(this);
        mb.add(m);
		m = new Menu("Help");
        m.add(new MenuItem("About"));
		m.addActionListener(this);
		mb.setHelpMenu(m);
        frame.setMenuBar(mb);
    }

    public void actionPerformed(ActionEvent evt) {
		if (evt.getSource() == filterText || evt.getSource() == filterChoice) 
			doCommand("Filter");
		else doCommand(evt.getActionCommand());
        //textPanel.doCommand(cmd); Protected method
    }
    
    void doCommand(String cmd) {
        if (cmd==null)
             return;
        if (cmd.equals("Reset"))
            allFields();
		else if (cmd.equals("Output"))
			output();
        else if (cmd.equals("Filter"))
            filter();
		else if (cmd.equals("Clear Filter"))
            filterText.setText("");	
        else if (cmd.equals("About"))
            about();
        else if (cmd.equals("Save As..."))
            textPanel.saveAs("");
        else if (cmd.equals("Cut"))
            {textPanel.copySelection();textPanel.clearSelection();}
        else if (cmd.equals("Copy"))
            textPanel.copySelection();
        else if (cmd.equals("Clear"))
            textPanel.clearSelection();
        else if (cmd.equals("Select All"))
            textPanel.selectAll();
        else if (cmd.equals("Copy All")) {
            textPanel.selectAll();
            textPanel.copySelection();
            textPanel.resetSelection();        
        } 
    }
	
    public void itemStateChanged(ItemEvent e) {
		if (e.getSource() == filterChoice) 
			doCommand("Filter");
	}

	void output() {
		String[] str_arr;
        String[] lines;
		String separator = "\t";
		String choice = separatorChoice.getSelectedItem();
		int output = outputChoice.getSelectedIndex();
		String outputString;
		if (choice.matches("Tab")) 
			separator = "\t";
		else if (choice.matches("Space")) 
			separator = " ";
		else if (choice.matches("Equals")) 
			separator = "=";	
			
        if (textPanel.getSelectionStart() == -1 || IJ.versionLessThan("1.35a")) {
            lines = textPanel.getText().split("\n");
        } else {
            int start = textPanel.getSelectionStart();
            int end = textPanel.getSelectionEnd();
            lines = new String[end - start + 1];
            for (int i = start; i <= end; i++) {
                lines[i - start] = textPanel.getLine(i);
            }
        }
		
        for (int i = 0; i < lines.length; i++) {
            str_arr = lines[i].trim().split("\t");
			if (str_arr.length == 2) 
				outputString = str_arr[0] + separator + str_arr[1] + "\n";
            else 
				outputString = str_arr[0] + "\n";
			switch (output) {
				case 0:{
					// Log
					IJ.log(outputString);
					break;
				} case 1: {
					// Terminal
					System.out.print(outputString);
				}
			}
        }
	}
	
    void filter() {
		String filterString = filterText.getText();
        if (textPanel == null || filterString.matches("")) 
            return;
		filterString=filterString.trim();
        textPanel.selectAll();
        textPanel.clearSelection();
        String[] data = allData.split("\n");
        int size = data.length;
        String line;
        String pattern = ".*";
        StringBuffer output = new StringBuffer();
		String filterItem = filterChoice.getSelectedItem();
        if (filterItem.matches("Starts With")) {
            pattern = new String("^" + filterString + ".*$");
        } else if (filterItem.matches("Field Contains")) {
            pattern = new String("^.*" + filterString  + ".*\t.*$");
        } else if (filterItem.matches("Value Contains")) {
            pattern = new String("^.*\t.*" + filterString  + ".*$");
        } else if (filterItem.matches("Ends With") ) {
            pattern = new String("^.*" + filterString  + "\t.*$");
        } else if (filterItem.matches("Regular Expression") ) {
            pattern = new String(filterString) ;
        }
        Pattern p = Pattern.compile(pattern);
        Matcher m;
        for (int i = 0; i < size; i++) {
            line = data[i];
            m = p.matcher(line);
            if (m.matches()) {
                output.append(line + "\n");
            } 
        }
        textPanel.append(output.toString());
    }
	
    void allFields() {
        if (textPanel == null) 
            return;
        textPanel.selectAll();
        textPanel.clearSelection();
        textPanel.append(allData);
        filterText.setText("");
    }
    
    /** 
    Returns this SearchWindow's MenuBar
    */ 
    public MenuBar getMenuBar() {
        return mb;
    }
    
    /** 
    Sets the MenuBar
    */
    public void updateMenuBar(MenuBar menuBar) {
        mb = menuBar;
        frame.setMenuBar(mb);
    }
    
    void about() {
        IJ.showMessage("About SearchWindow", aboutText);
    }
    
    public String getAboutText() {
        return aboutText;
    }

    /**
    Adds one or lines of text to the window.
    @param text        The text to be appended. Multiple
                    lines should be separated by \n.
    */
    public void append(String text) {
        textPanel.append(text);
    }
    
    /** Set the font that will be used to display the text. */
    public void setFont(Font font) {
        textPanel.setFont(font);
    }
  
    boolean openFile(String path) {
        OpenDialog od = new OpenDialog("Open Text File...", path);
        String directory = od.getDirectory();
        String name = od.getFileName();
        if (name==null)
            return false;
        path = directory + name;
        
        IJ.showStatus("Opening: " + path);
        try {
            BufferedReader r = new BufferedReader(new FileReader(directory + name));
            load(r);
            r.close();
        }
        catch (Exception e) {
            IJ.error(e.getMessage());
            return true;
        }
        textPanel.setTitle(name);
        frame.setTitle(name);
        IJ.showStatus("");
        return true;
    }
    
    /** Returns a reference to this SearchWindow's TextPanel. */
    public TextPanel getTextPanel() {
        return textPanel;
    }
    /** Returns a reference to this SearchWindow's UtilityPanel. */
    public Panel getUtilityPanel() {
        return utilityPanel;
    }

    /** Appends the text in the specified file to the end of this SearchWindow. */
    public void load(BufferedReader in) throws IOException {
        int count=0;
        while (true) {
            String s=in.readLine();
            if (s==null) break;
            textPanel.appendLine(s);
        }
    }
    
    public void close() {
        frame.setVisible(false);
		frame.dispose();
        WindowManager.removeWindow(frame);
        //textPanel.flush(); protected method in TextPanel - replaced with
        textPanel.selectAll();
        textPanel.clearSelection();
        textPanel = null;
  		// Under most circumstances the JVM terminates when the last window is closed, however occasionally the JVM fails to exit
		// see http://java.sun.com/j2se/1.5.0/docs/api/java/awt/doc-files/AWTThreadIssues.html     
		// In order to ensure jvm exits when this window is closed and this window was created by a command line plugin:
		if (IJ.getInstance() == null) {
			System.exit(0);
		}
    }
    
    public void focusGained(FocusEvent e) {
        WindowManager.setWindow(frame);
    }

    public void focusLost(FocusEvent e) {}
    
    public  void windowActivated(WindowEvent e) {
         WindowManager.setWindow(frame);
     }
    
    public void windowClosed(WindowEvent e) {}
        //Invoked when a window has been closed as the result of calling dispose on the window.
    public void windowClosing(WindowEvent e) {
        //Invoked when the user attempts to close the window from the window's system menu.
        close(); 
    }
    public void windowDeactivated(WindowEvent e) {
            //Invoked when a Window is no longer the active Window.
    }
    public void windowDeiconified(WindowEvent e) {
            //Invoked when a window is changed from a minimized to a normal state.
    }
    public void windowIconified(WindowEvent e) {
            //Invoked when a window is changed from a normal to a minimized state.
    }
    public void windowOpened(WindowEvent e)  {
        //Invoked the first time a window is made visible.
    }
}
