import ij.*;
import ij.gui.*;
import ij.io.*;
import java.awt.*;
import java.io.*;
import java.lang.*;
import ij.plugin.*;
import ij.text.*;
import com.xinapse.util.*;
import java.util.*;
import java.util.regex.*;
import java.awt.event.*;
import com.xinapse.loadableimage.*;
import com.xinapse.multisliceimage.*;
import com.xinapse.multisliceimage.UNC.*;

 /** Author: Jon Jackson (j.jackson at ucl.ac.uk)
 *
 *   History:  28/09/05 Last Modified
 *
 *   Description:
 *   UNC Header plugin.  Displays a UNC header fields in a SearchWindow
 *	 (University of North Carolina (UNC) libim format modified to include image information fields by 
 *	 David Plummer, Medical Physics and Bioengineering, University College London)
 *
 *	 This plugin requires the xinapse.jar classes from www.xinapse.com
 *
 */

public class UNC_Header  implements PlugIn, ItemListener  {
    InfoList header;
	UNCImage unc;
	ImagePlus imp;
    SearchWindow window;
    TextPanel textPanel;
    String headerString;
    String[] filters;
    Choice fieldsChoice;
	Choice sliceChoice;
	Choice infoChoice;
	int slices = 1;
	String originalFilePath;
	boolean commandline=false;
	String filterFname="UNCHeaderFilters.txt";
	String newFilterMsg="To add a new filter, create or edit UNC_Header_Filters.txt\n" +
				"in the ImageJ plugins directory. Entries are comma separated \n" +
				"with each filter on a single line. The first entry is the filter name\n" +
				"and all subsequent entries are fields included by this filter";
    
	// This method is called when executed from commandline
	public static void main(String args[]) {
        if (args.length<1) {
            IJ.log("usage: UNC_Header <image>");
			System.exit(0);
		}
        UNC_Header unchdr = new UNC_Header();
		if (unchdr.open(args[0]) == false) {
			System.exit(0);
		}
	}
	
	public boolean open(String filename) {
		originalFilePath = filename;
		if (! openUNC(filename)) {
			return false;
		}
		header = unc.getInfoList();
		if (header == null) {
            IJ.error("Image does not have an UNC header");
            return false;
        }
		try {
			slices = unc.getNSlices();
		} catch (MultiSliceImageException msie) {} 
		headerString = filterEmptyFields();    
        if (headerString == null) 
            return false;
		File file = new File(filename);
        window = new SearchWindow(file.getName() + " Header", "Field\tValue", headerString, 480, 650); 
        textPanel = window.getTextPanel();
        if (IJ.isJava14()) {
            addChoices();
        } else {
            IJ.showStatus("Search facility requires Java 1.4");
        }
        return true;
	}
	
	boolean openUNC(String filename) {
		File file = new File(filename);
		if (! file.exists()) {
			System.out.print("Cannot find " + filename + "\n");
			return false; 
		}
		try {
			unc = new UNCImage(file,"r");
		} catch (UNCException unce) {
			System.out.print(unce.getMessage());
			return false;
		} catch (FileNotFoundException fnfe) {
			System.out.print(fnfe.getMessage());
			return false;
		} 
		return true;
    }
	
	// This method called when executed inside imagej
    public void run(String arg) {
    
        imp = WindowManager.getCurrentImage();
        if (imp == null) {
            IJ.noImage(); 
            return;
        }    
        header = (InfoList)imp.getProperty("InfoList");
        if (header == null) {
            IJ.error("Image does not have an UNC header");
            return;
        }   
        headerString = filterEmptyFields();    
        if (headerString == null) 
            return;
		slices = imp.getNSlices();
		FileInfo fInfo = imp.getOriginalFileInfo();
		originalFilePath = fInfo.directory + fInfo.fileName;
        window = new SearchWindow(imp.getTitle() + " Header", "Field\tValue", headerString, 480, 650); 
        textPanel = window.getTextPanel();
        addChoices();
    }
    
    String filterEmptyFields() {
	
        if (header == null) 
            return null;
        int size = header.size();
        String[] str_arr;
        InfoList headerFiltered = new InfoList();
        StringBuffer headerBuffer = new StringBuffer();
        Pattern p1 = Pattern.compile("echo_[0-9]_[abwl].*");
        Matcher m1, m2, m3;
        Pattern p2 = Pattern.compile(".*[Uu]nsupported.*");
        Pattern p3 = Pattern.compile("UserData[0-9].*");
        
        for (int i = 0; i < size; i++) {
            str_arr = header.get(i).toString().split("=");
            if (str_arr.length < 2) {
                continue;
            }
            m1 = p1.matcher(str_arr[0]);
            if (m1.matches() && str_arr[1].startsWith("0")) {
                continue;
            }
            m2 = p2.matcher(str_arr[1]);
            if (m2.matches()) {
                continue;
            }
            m3 = p3.matcher(str_arr[0]);
            if (m3.matches() && str_arr[1].startsWith("0")) {
                continue;
            }
            if (str_arr[1].startsWith("(not found in image header)")) {
                continue;
            }
            if (str_arr[0].startsWith("user_defined_variable") && str_arr[1].startsWith("0")) {
                continue;
            }
            if (str_arr[1].startsWith("(no value)")) {
                continue;
            }
            
            // if header value does not match one of the empty field templates above then add to the infolist
            try {
                headerFiltered.putInfo(str_arr[0], str_arr[1]);
                headerBuffer.append(str_arr[0] + "\t" + str_arr[1] + "\n");
            } catch (InfoListException ile) {}
            
        }
        header = headerFiltered;
        return headerBuffer.toString();
    }
    
    void filter(int index) {
        if (header == null || textPanel == null || filters[index] == null) 
            return;
        StringBuffer headerBuffer = new StringBuffer();
        String entry;
        String key;
        String[] paramList = filters[index].split(",");
        for (int i = 1; i < paramList.length; i++) {
            key = paramList[i].trim();
            try {
                entry = header.getInfo(key);
                headerBuffer.append(key + "\t" + entry + "\n");
            } catch (InfoNotFoundException infe) {
                //IJ.log(infe.getMessage());
            }
        }
        window.setText(headerBuffer.toString());
    }
    
    void allFields() {
        if (headerString != null) 
            window.setText(headerString);
    }
	
	void sliceInfo(int n) {
		if (unc == null) {
			openUNC(originalFilePath);
		}
		//window.setText(unc.getInfoList(0,n).toString().replace(" ","").replace("=", "\t"));
        window.setText(unc.getInfoList(0,n).toString().trim().replace('=', '\u0009'));
	}	
	
    void addChoices() {
        // Add Output to Terminal Button
        ArrayList list = new ArrayList();
		Panel buttonPanel = (Panel)window.getUtilityPanel().getComponent(0);
		String pluginsDir = IJ.getDirectory("plugins");
		if (pluginsDir == null) 
			// Case where ImageJ is opened from the command line with -Dplugins.dir=xxx
			pluginsDir = System.getProperty("plugins.dir") + File.separator + "plugins" + File.separator;
        try {
            BufferedReader r = new BufferedReader(new FileReader(pluginsDir + filterFname));
            int count=0;
            while (true) {
                String s=r.readLine();
                if (s==null) break;
                list.add(s);
            }
            r.close();
        }
        catch (Exception e) {
            IJ.showStatus("Filters file not installed");
        }
		fieldsChoice = new Choice();
        if (list.size() != 0) {
        	fieldsChoice.add("All Fields");
        	filters = new String[list.size()];
        	String temp;
        	int index = 0;
        	filters = (String[])list.toArray(filters);
        	for (int i = 0; i < filters.length; i++) {
            	temp = filters[i];
            	if (temp != null) {
                	index = temp.indexOf(",");
                	temp = temp.substring(0,index);
                	fieldsChoice.add(temp);
            	}
        	}
		}
        fieldsChoice.add("New Filter");
        fieldsChoice.addItemListener(this);
        buttonPanel.add(fieldsChoice, 0);
		sliceChoice = new Choice();
		for (int i = 1; i <= slices; i++) {
			sliceChoice.add(String.valueOf(i));
		}
		sliceChoice.setEnabled(false);
		sliceChoice.addItemListener(this);
		buttonPanel.add(sliceChoice, 0);
		infoChoice = new Choice();
		infoChoice.add("Global");
		infoChoice.add("Slice");
		infoChoice.addItemListener(this);
		buttonPanel.add(infoChoice, 0);
        buttonPanel.validate();
    }
    
	    
    public void itemStateChanged(ItemEvent e) {
        Object source = e.getSource();
		if (source == fieldsChoice) {
			String field = fieldsChoice.getSelectedItem();
        	if (field.equals("All Fields")) {
            	allFields();
        	} else if (field.equals("New Filter")) {
            	IJ.showMessage(newFilterMsg);
        	} else {
            	int index = fieldsChoice.getSelectedIndex();
            	filter(index - 1);
        	}
		} else if (source == infoChoice) {
			String field = infoChoice.getSelectedItem();
			if (field.equals("Global")) {
				sliceChoice.setEnabled(false);
				fieldsChoice.setEnabled(true);
				if (fieldsChoice.getSelectedItem().equals("All Fields")) {
            		allFields();
        		} else {
            		int index = fieldsChoice.getSelectedIndex();
            		filter(index - 1);
        		}
			} else if (field.equals("Slice")) {
				fieldsChoice.setEnabled(false);
				sliceChoice.setEnabled(true);
				sliceInfo(sliceChoice.getSelectedIndex());
			}
		} else if (source == sliceChoice) {	
			sliceInfo(sliceChoice.getSelectedIndex());
		}
    }
	
}

