import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.io.*;
import ij.io.*;
import ij.plugin.*;
import ij.measure.*;
import ij.text.*;
import com.xinapse.loadableimage.*;
import com.xinapse.multisliceimage.*;
import com.xinapse.multisliceimage.UNC.*;
import com.xinapse.util.*;

 /** Author: Jon Jackson (j.jackson at ucl.ac.uk)
 *
 *   History:  3/2/05   Added header file support
 *             12/11/05 Added getImagePlus() method as a workaround - really needs to be re-written extending ImagePlus
 *             14/06/06 Extends ImagePlus
 *
 *   Description:
 *   UNC reader plugin.  Reads images in UNC image file format
 *	 (University of North Carolina (UNC) libim format modified to include image information fields by 
 *	 David Plummer, Medical Physics and Bioengineering, University College London)
 *   UNC is the main format used by BAMM (http://www-bmu.psychiatry.cam.ac.uk/software/) fMRI processing software
 *
 *	 This plugin requires the xinapse.jar classes from www.xinapse.com 
 *       (Download and install the Jim image processing package from www.xinapse.com (there is no fee during the trial period))
 *       (Copy the xinapse.jar file from the Jim package into the plugins folder of your ImageJ directory)
 *
 */

public class UNC_Reader extends ImagePlus implements PlugIn {
	
    public void run(String arg) {
        UNCFileFilter unc_filter = new UNCFileFilter();
        
        OpenDialog od = new OpenDialog("Open UNC...", arg);
        String directory = od.getDirectory();
        String name = od.getFileName();
        if (name==null) {
            // user cancelled
              return;
        }
        if (directory == "") {
            directory = System.getProperty("user.dir") + File.separator;
            if (directory == null) directory = "";
        }
        File file_obj = new File(directory, name);
        if (! file_obj.exists()) {
            IJ.error("File does not exist");
            return;
        }
        if (! unc_filter.accept(file_obj)) {
            IJ.error("File does not have UNC signature");
            return;
        }
        IJ.showStatus("Opening: " + directory + name);
        load(directory, name);
        if (arg.equals("")) {
            updateAndDraw();
            show();
            updateAndDraw();
        }
    }
    
    public void load(String directory, String name) {
        FileInfo fi_ = new FileInfo();
        UNCImage unc;
        ImageProcessor ip_=null;
        ImageStack stack_=null;
        Calibration cal_;
        int width_, height_, nSlices_;
        Double min_unc=null;
        Double max_unc=null;
        UNCPixFormat data_type;
        Object pixels=null;
        boolean toSigned=false;
        double max, min;
        if ((name == null) || (name == "")) return;
        if (!directory.endsWith(File.separator) && directory !="") directory += File.separator;
        
        fi_.fileName = name;
        fi_.directory = directory;
        
        IJ.showStatus("Loading UNC File: " + directory + name);
 
        try {
            unc =  new UNCImage(directory + name,"r");
        } catch(FileNotFoundException fnf)  {
            IJ.error("File not found: " + fnf.getMessage());
            return;
        } catch(UNCException uncex) {
              IJ.error("Error reading UNC file: " + uncex.getMessage());
               return;
           }
        try {    
            data_type = UNCPixFormat.getFormat(unc.getPixelDataType());
            
        } catch (UNCException uncex) {
            IJ.error("Error determining pixel format: " + uncex.getMessage());
               return;
        }
        try {
            width_ = unc.getNCols();
            height_ = unc.getNRows();
            nSlices_ = unc.getNSlices();
        } catch (MultiSliceImageException msie) {
            IJ.error("Error getting image dimensions: " + msie.getMessage());
            return;
        }
        
        if (data_type.equals(UNCPixFormat.GREY)) {
            ip_ = new ShortProcessor(width_, height_);
            fi_.fileType = FileInfo.GRAY16_SIGNED;
            toSigned = true;
        } else if (data_type.equals(UNCPixFormat.SHORT)) {
            ip_ = new ShortProcessor(width_, height_);
            fi_.fileType = FileInfo.GRAY16_SIGNED;
            toSigned = true;
        } else if (data_type.equals(UNCPixFormat.BYTE)) {
            ip_ = new ByteProcessor(width_, height_);
            fi_.fileType = FileInfo.GRAY8;
        } else if (data_type.equals(UNCPixFormat.REAL)) {
            ip_ = new FloatProcessor(width_, height_);
            fi_.fileType = FileInfo.GRAY32_FLOAT;
        } else {
            IJ.error("UNC Reader cannot currently support UNC pixel type " + data_type.toString() + " \nSupported types are BYTE, SHORT, GREY, and REAL");
            return;
        }
        
        IJ.showStatus("Reading image data...");
        
        if (nSlices_ == 1) {
        // read image data for 2D image
            try {
                pixels = unc.getSlice(0);
                if (toSigned) {
                    to16BitSigned((short[])pixels);
                }
            } catch (InvalidImageException iie) {
                IJ.log(iie.getMessage());
                return;
            } catch (OutOfMemoryError e) {
                unc = null;
                pixels = null;
                IJ.showProgress(1.1);
                IJ.outOfMemory("UNC Reader Plugin");
                return;
            }
            ip_.setPixels(pixels);
            setProcessor(name, ip_);
        }    else    {
        // read image data for 3D image
            stack_ = new ImageStack(width_, height_, null);
            try {
                for (int i=0; i<nSlices_; i++)  {
                    IJ.showStatus("Reading: " + i + "/" + nSlices_);
                    pixels = unc.getSlice(i);
                    if (toSigned) {
                        to16BitSigned((short[])pixels);
                    }
                    stack_.addSlice(null, pixels);
                    IJ.showProgress(i, nSlices_);
                }
            } catch (InvalidImageException iie) {
                IJ.log("Error reading stack " + iie.getMessage());
                return;
            } catch (OutOfMemoryError e) {
                unc = null;
                stack_ = null;
                IJ.showProgress(1.1);
                IJ.outOfMemory("UNC Reader Plugin");
                return;
            }
            setStack(name, stack_);
        }
        
        IJ.showStatus("Getting image info...");
        
        fi_.width = width_;
        fi_.height = height_;
        fi_.nImages = nSlices_;
        
        cal_ = new Calibration(this);
        cal_.setUnit("mm"); // mm is default units returned from unc.getInfo

        try {
            cal_.pixelWidth = unc.getPixelXSize();
            cal_.pixelHeight = unc.getPixelYSize();
            cal_.pixelDepth = unc.getPixelZSize();
        } catch (ParameterNotSetException pnse) {
            // default pixel size 1*1*1 inherited from Calibration
            IJ.log("Pixel Size not set");
        }        
        
        if (toSigned) {
            double[] coeff = new double[2];
            coeff[0] = -32768.0;
            coeff[1] = 1.0;
            cal_.setFunction(Calibration.STRAIGHT_LINE, coeff, "gray value");
        }            
        
        try {
            // get min and max pixel value from unc header
            min_unc = unc.getMin();
            max_unc = unc.getMax();
        } catch (InvalidImageException iie) {
        }        
        
        if (min_unc != null && max_unc != null) {
            // set display max and min according to unc header
            min = min_unc.doubleValue();
            max = max_unc.doubleValue();
            if (toSigned) {
                min += 32768.0;
                max += 32768.0;
            }            
            this.getProcessor().setMinAndMax(min, max);
        } else {
            // if it wasn't found from the header, calculate from pixel data
            min = Double.MAX_VALUE;
            max = -Double.MAX_VALUE;    
            if(nSlices_ == 1) {
                // calculate for single image
                this.getProcessor().resetMinAndMax();
            } else {
                // calculate for stack
                for (int i=1; i<=nSlices_; i++) {
                    IJ.showStatus("Calculating stack_ min and max: "+i+"/"+nSlices_);
                    ip_ = stack_.getProcessor(i);
                    ip_.resetMinAndMax();
                    if (ip_.getMin()<min)
                        min = ip_.getMin();
                    if (ip_.getMax()>max)
                        max = ip_.getMax();
                }
            }
            this.getProcessor().setMinAndMax(min, max);
        }

        InfoList list = unc.getInfoList();
        if (list == null) {
            IJ.log("Could not read UNC header fields");
        }
        
        setProperty("InfoList", list);
        setFileInfo(fi_);
        setCalibration(cal_); 
        
        try {        
            unc.close();
        } catch (MultiSliceImageException msie) {
            IJ.log("Error closing image: " + msie.getMessage());
        }        
        IJ.showProgress(1.1);
        return;                
    }
    
    void to16BitSigned(short[] pixels) {
        for (int i = 0; i < pixels.length; i++) {
            pixels[i] += 32768;
        }
        return;
    }
	
}
