import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.io.*;
import java.lang.*;
import ij.io.*;
import ij.plugin.*;
import ij.measure.*;
import com.xinapse.loadableimage.*;
import com.xinapse.multisliceimage.*;
import com.xinapse.multisliceimage.UNC.*;
import com.xinapse.util.*;

 /** Author: Jon Jackson (j.jackson at ucl.ac.uk)
 *
 *   History:  3/2/05 Added header file support
 *
 *   Description:
 *   UNC writer plugin.  Saves in UNC image file format
 *	 (University of North Carolina (UNC) libim format modified to include image information fields by 
 *	 David Plummer, Medical Physics and Bioengineering, University College London)
 *   UNC is the main format used by BAMM (http://www-bmu.psychiatry.cam.ac.uk/software/) fMRI processing software
 *
 *	 This plugin requires the xinapse.jar classes from www.xinapse.com
 *
 */

public final class UNC_Writer implements PlugIn {

    public void run(String arg) {
        ImagePlus imp = WindowManager.getCurrentImage();
         if (imp==null) {
            IJ.noImage(); 
            return; 
        }
        String directory = "", name = ""; 
        if ((arg==null) || (arg=="")) {
            SaveDialog sd = new SaveDialog("Save as UNC", imp.getTitle(), "");
            directory = sd.getDirectory();
            name = sd.getFileName();
        } else {
            File file = new File(arg); 
            if (file.isDirectory()) {
                directory = arg;
                name = imp.getTitle();
            } else {
                directory = file.getParent(); 
                name = file.getName(); 
            }    
        }
        if (name == null || name == "") {
            return;
        } else if (directory == "") {
            directory = System.getProperty("user.dir") + File.separator ;
            if (directory == null) directory = "";
        }
        File file = new File(directory, name);
		// JFileChooser has overwrite check
        /*if (file.isFile()) {
            if (ij.macro.Interpreter.isBatchMode()) {
                IJ.write(directory + name + " Exists, no changes made");
                return;
            } else {
                Frame parent = new Frame();
                YesNoCancelDialog overwrite = new YesNoCancelDialog(parent, "", name + " exists, overwrite?");
                if (! overwrite.yesPressed())
                    return;
            }
        } */
        boolean write_header = false;
        if (imp.getProperty("InfoList") == null) {
            write_header = false;
        } else {
            if (ij.macro.Interpreter.isBatchMode()) {
                write_header = false;
            } else {
                Frame parent = new Frame();
                YesNoCancelDialog header = new YesNoCancelDialog(parent, "", "Original image contains UNC header fields\nSave header fields with new image?");
                if (header.yesPressed())
                    write_header = true;
                else if (header.cancelPressed())
                    return;
            }
        }
        save(imp, directory, name, write_header);
        IJ.showStatus(""); 
    }
    
    public void save(ImagePlus imp, String directory, String name, boolean write_header) {
        if (name == null) return;
        if (!directory.endsWith(File.separator) && directory !="") directory += File.separator; 
        
        int width, height, slices;
        int dimc;
        int[] dimv;        
        ImageProcessor ip = imp.getProcessor();
        Calibration cal = imp.getCalibration();
        int type = imp.getType();
        UNCImage unc;
        Object pixels;
        boolean fromSigned = cal.isSigned16Bit();
        slices = imp.getStackSize();
        
        IJ.showStatus("Saving as UNC: " + directory + name);
        
        slices = imp.getStackSize();
        width = imp.getWidth();
        height = imp.getHeight();
        if (slices == 1) {
            // for dimv, element 0 is the most slowly varying element
            dimc = 2;
            dimv = new int[2];
            dimv[1] = width;
            dimv[0] = height;
        } else if (slices > 1) {
            dimc = 3;
            dimv = new int[3];
            dimv[2] = width;
            dimv[1] = height;  
            dimv[0] = slices;    
        } else {
            IJ.error("Unexpected Stack size: " + slices);
            return;
        }

        try {
            if (type == ImagePlus.GRAY8) {
                unc = new UNCImage(UNCPixFormat.BYTE, dimc, dimv);
            } else if (type == ImagePlus.GRAY16) {
                unc = new UNCImage(UNCPixFormat.GREY, dimc, dimv);
            } else if (type == ImagePlus.GRAY32) {
                unc = new UNCImage(UNCPixFormat.REAL, dimc, dimv);
            } else {
                IJ.error("UNC writer presently supports only GRAY8, GRAY16 and GREY32 data types");
                return;
            }
        } catch (UNCException unce) {
            IJ.error("Error creating UNC file: " + unce.getMessage());
            return;
        }
        
        if (write_header) {
            InfoList header = (InfoList)imp.getProperty("InfoList");
            if (header != null) {
                unc.setInfoList(header);
            }
        }

        if (slices == 1) {
            try {
                // UNCImage copies pixel values from supplied array to pixel data
                // cf. ImageProcessor gives a reference to the pixel array
                ip = imp.getProcessor();
                unc.putSlice(ip.getPixels(), 0);
                if (fromSigned) {
                    try {
                    pixels = unc.getSlice(0);
                    from16BitSigned((short[])pixels);
                    } catch (InvalidImageException iie) {
                        IJ.error(iie.getMessage());
                        return;
                    }
                    unc.putSlice(pixels, 0);
                } 
            } catch (MultiSliceImageException msie) {
                IJ.error("Error with image data: " + msie.getMessage());
                return;
            }

        } else {
            ImageStack stack = imp.getStack();
            try {
                for (int i = 1; i <= slices; i++) {
                    unc.putSlice(stack.getPixels(i), i - 1);
                    if (fromSigned) {
                        pixels = unc.getSlice(i - 1);
                        from16BitSigned((short[])pixels);
                        unc.putSlice(pixels, i - 1);
                    }
                
                }    
            } catch (MultiSliceImageException msie) {
                IJ.error("Error with stack data: " + msie.getMessage());
                return;
            } catch (InvalidImageException iie) {
                IJ.error(iie.getMessage());
                return;
            }
        }

        try {
            unc.setPixelXSize((float)cal.pixelWidth);
            unc.setPixelYSize((float)cal.pixelHeight);
            unc.setPixelZSize((float)cal.pixelDepth);
            unc.setTitle(imp.getTitle());
        } catch (MultiSliceImageException msie) {
        }

        
        File file = new File(directory, name);
        if (file.isFile()) {
            // unc.write throws error if file exists.
            if (! file.delete()) {
                IJ.write("Could not overwrite file " + directory + name);
                return;
            }
        }
        
        try {
            unc.write(directory + name);
        } catch (MultiSliceImageException msie) {
            IJ.error("Error creating file: " + msie.getMessage());
            return;
        }
    }
    
    private void from16BitSigned(short[] pixels) {
        for (int i = 0; i < pixels.length; i++) {
            pixels[i] -= 32768;
        }
        return;
    }
}    
